from sqlalchemy import create_engine, MetaData, event, Table
from sqlalchemy.orm import Session
from sqlalchemy.ext.automap import automap_base
from sqlalchemy_utils.functions.database import database_exists, create_database
import sys
from madmigration.utils.helpers import issue_url,app_name,parse_uri

@event.listens_for(Table, "after_parent_attach")
def before_parent_attach(target, parent):
    if not target.primary_key and "id" in target.c:
        print(target)

def goodby_message(message, exit_code=0):
    print(message, flush=True)
    sys.exit(int(exit_code))

class SourceDB:
    def __init__(self, source_uri):
        if not database_exists(source_uri):
            goodby_message(database_not_exists(source_uri), 0)
        self.base = automap_base()
        self.engine = create_engine(source_uri, echo=False)
        self.base.prepare(self.engine, reflect=True)
        self.session = Session(self.engine, autocommit=False, autoflush=False)
        



class DestinationDB:
    def __init__(self, destination_uri):
        if not database_exists(destination_uri):
            while True:
                database = parse_uri(destination_uri)
                msg = input(f"The database {database} does not exists, would you like to create it in the destination?(y/n) ")
                if msg.lower() == "y":
                    try:
                        create_database(destination_uri)
                        sys.stdout.write("Database created ..")
                        break
                    except Exception as err:
                        goodby_message(database_not_exists(err), 1)
                    break
                elif msg.lower() == "n":
                    goodby_message("Destination database does not exit \nExiting ..", 0)
                    break
                print("Please, select command")

        
        self.base = automap_base()
        self.engine = create_engine(destination_uri)
        # self.base.prepare(self.engine, reflect=True)
        self.session = Session(self.engine, autocommit=False, autoflush=False)


def database_not_exists(database):
    
    """This function will be executed if there is no database exists """

    database = parse_uri(database)

    usage = [
        "",
        f"😭 Error: Source database '{database}'  does not exists",
        "",
        f"Run '{app_name()} --help' for usage.",
        "",
        f"🥳  if you think something is wrong please feel free to open issues 👉'{issue_url()}'👈 ",
        "",
        "Exiting ...",
        ""
    ]
    return "\n".join(usage)


import pytest
from madmigration.config.conf import Config
from madmigration.utils.helpers import  get_cast_type, get_column_type, detect_driver
from sqlalchemy import (
    Integer,
    String,
    DateTime,
    Date,
    BigInteger,
    VARCHAR,
    Float,
    TIMESTAMP
)
from datetime import  datetime
from madmigration.mysqldb.migration import Migrate as mysql_migrate
from madmigration.postgresqldb.migration import Migrate as postgres_migrate


def get_cast_type_test(temp_data):

    assert get_cast_type(temp_data[0].lower()) == str
    assert get_cast_type(temp_data[1]) == str
    assert get_cast_type(temp_data[2]) == int
    assert get_cast_type(temp_data[3]) == int
    assert get_cast_type(temp_data[5]) == float
    assert get_cast_type(temp_data[6]) == datetime


def get_column_type_test(temp_data):

    assert get_column_type(temp_data[1]) == String
    assert get_column_type(temp_data[3].lower()) == Integer
    assert get_column_type(temp_data[4]) == BigInteger
    assert get_column_type(temp_data[5]) == Float
    assert get_column_type(temp_data[6]) == DateTime
    assert get_column_type(temp_data[8]) == TIMESTAMP
    assert get_column_type(temp_data[9]) == VARCHAR


def detect_driver_test():
      
    assert detect_driver("mysqldb") == mysql_migrate
    assert detect_driver("psycopg2") == postgres_migrate






import pytest

# Wherner we need to run some code before test, we use pytest fixture;
# Modular fixtures for managing small or parametrized long-lived test resources;
# Pytest is a feature-rich, plugin-based ecosystem for testing your Python code;


@pytest.fixture
def temp_data():
    temp = [
        'str',
        'string',
        'int',
        'integer',
        'biginteger',
        'float',
        'datetime',
        'date',
        'timestamp',
        'varchar']
    return temp



from pydantic import BaseModel, validator
from typing import Union, List, Any, AnyStr, Dict


class ForeignKeySchema(BaseModel):
    use_alter: bool = True
    table_name: str
    column_name: str
    ondelete: str = "CASCADE"
    onupdate: str = "NO ACTION"
    onupdate: str = "NO ACTION"

class OptionsSchema(BaseModel):
    primary_key: bool = False
    nullable: bool = False
    default: Any = None
    index: Union[bool,None] = None
    unique: Union[bool,None] = None
    autoincrement: bool = False
    foreign_key: Union[ForeignKeySchema,None] = None
    length: Union[int,None] = None
    type_cast: Union[str,None] = None

    # @validator("type_cast")   #TODO check types
    # def validate_cast(cls,v,values): 
    #     if v == "integer" and values["length"] != None:
    #         raise ValueError(f"Type {v} has no length")
    #     return v


class SourceConfigSchema(BaseModel):
    SourceConfig: Dict[str, str]


class DestinationConfigSchema(BaseModel):
    DestinationConfig: Dict[str, str]


class SourceTableSchema(BaseModel):
    name: str
    

class DestTableSchema(BaseModel):
    name: str
    create: bool = False
    
class DestinationColumn(BaseModel):
    name: str
    options: OptionsSchema

class SourceColumn(BaseModel):
    name: str


class ColumnParametersSchema(BaseModel):
    destinationColumn: DestinationColumn
    sourceColumn: SourceColumn


class TablesInfo(BaseModel):
    SourceTable: SourceTableSchema
    DestinationTable: DestTableSchema
    MigrationColumns: List[ColumnParametersSchema]


class MigrationTablesSchema(BaseModel):
    migrationTable: TablesInfo


class ConfigSchema(BaseModel):
    Configs: List[Union[SourceConfigSchema, DestinationConfigSchema]]
    migrationTables: List[MigrationTablesSchema]
    version: float



from sqlalchemy.schema import DropConstraint, DropTable
from sqlalchemy.engine import reflection
from madmigration.errors import TableExists
from alembic.migration import MigrationContext
from alembic.operations import Operations
from sqlalchemy import DateTime
from sqlalchemy.dialects.mysql import (
    VARCHAR,
    INTEGER,
    NVARCHAR,
    SMALLINT,
    SET,
    BIGINT,
    BINARY,
    BOOLEAN,
    CHAR,
    DATE,
    DATETIME,
    DECIMAL,
    ENUM,
    FLOAT,
    JSON,
    NUMERIC,
    TEXT,
)
from alembic import op
from collections import defaultdict
from sqlalchemy_utils import UUIDType
from madmigration.config.conf import Config
from madmigration.db_operations.operations import DbOperations
from madmigration.basemigration.base import BaseMigrate
from pprint import pprint

class MysqlMigrate(BaseMigrate): 
    def __init__(self, config: Config,destination_db):
        super().__init__(config,destination_db)
        self.dest_fk = defaultdict(list)
        self.db_operations = MysqlDbOperations(self.engine)
        self._drop_tables = []
        self.collect_table_names()
        self.collect_drop_fk()
    

    def collect_drop_fk(self):
        """ 
        Collect foreign key constraints for tables so 
        that if something goes wrong, delete everything
        """
        try:
            conn = self.engine.connect()
            transactional = conn.begin()
            inspector = reflection.Inspector.from_engine(self.engine)

            for table_name in inspector.get_table_names():
                if table_name in self.table_list:
                    for fk in inspector.get_foreign_keys(table_name):
                        if not fk["name"]:
                            continue
                        self.dest_fk[fk["referred_table"]].append((table_name,fk["name"]))
                        self.contraints_columns[table_name].add(*fk["constrained_columns"])
            transactional.commit()
        except Exception as err:
            print("collect_drop_fk own -> ",err)
            return False
        finally:
            conn.close()

    def get_input(self,table_name):
        while True:
            answ = input(
                f"Table with name '{table_name}' already exist,\
'{table_name}' table will be dropped and recreated,your table data will be lost,process?(y/n) ")
            if answ.lower() == "y":
                if self.dest_fk[table_name]:
                    self.db_operations.drop_fk(self.dest_fk[table_name])
                self._drop_tables.append(table_name)
                return False
            elif answ.lower() == "n":
                return True
            else:
                continue

    def check_table(self, table_name: str) -> bool:
        """ Check table exist or not, and wait user input """
        try:
            if self.engine.dialect.has_table(self.engine.connect(), table_name):
                return self.get_input(table_name)
            return False
        except Exception as err:
            print("check_table -> ",err)
            return False

    def process(self):
        """
        Create and check existing tables. 
        Collect foreign key constraints
        """
        try:
            if self._drop_tables:
                self.db_operations.bulk_drop_tables(*self._drop_tables)
            self.update_table()
            self.create_tables()
            self.db_operations.create_fk_constraint(self.fk_constraints,self.contraints_columns)
            return True
        except Exception as err:
            print("create_tables err -> ", err)

    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type
        :param type_name: str
        :return: object class
        """
        return {
            "string": VARCHAR,
            "varchar": VARCHAR,
            "integer": INTEGER,
            "nvarchar": NVARCHAR,
            "smallint": SMALLINT,
            "set": SET,
            "bigint": BIGINT,
            "binary": BINARY,
            "boolean": BOOLEAN,
            "bool": BOOLEAN,
            "char": CHAR,
            "date": DATE,
            "datetime": DATETIME,
            "decimal": DECIMAL,
            "enum": ENUM,
            "float": FLOAT,
            "json": JSON,
            "numeric": NUMERIC,
            "text": TEXT,
            "uuid": CHAR(36),
        }.get(type_name.lower())


class MysqlDbOperations(DbOperations):
    def drop_fk(self,fk_constraints: str):
        try:
            conn = self.engine.connect()
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            # [('todo', 'todo_ibfk_1')]
            for fk in fk_constraints:
                op.drop_constraint(fk[1], fk[0], type_="foreignkey")
        except Exception as err:
            print("fk_drop -> ",err)
        finally:
            conn.close()


from sqlalchemy.schema import DropConstraint, DropTable
from sqlalchemy import MetaData, ForeignKeyConstraint, Table
from alembic.migration import MigrationContext
from sqlalchemy.engine import reflection
from alembic.operations import Operations
from alembic import op


class DbOperations:
    def __init__(self,engine):
        self.engine = engine
        self.metadata = MetaData()

    def drop_table(self, table_name):
        """ Drop table with given name """
        try:
            conn = self.engine.connect()
            # context config for alembic
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            op.drop_table(table_name)
            return True
        except Exception as err:
            print("drop_tables err -> ",err)
        finally:
            conn.close()


    def bulk_drop_tables(self, *table_name):
        """ Drop table with given name """
        try:
            conn = self.engine.connect()
            # context config for alembic
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            for tb in table_name:
                op.drop_table(tb)
            return True
        except Exception as err:
            print("bulk_drop_tables err -> ",err)
        finally:
            conn.close()

    def update_column(self,table_name, column_name, col_type, **options):
        """ Updated existing table column with new column """
        try:
            conn = self.engine.connect()
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)

            op.alter_column(table_name, column_name,type_=col_type,postgresql_using=f"{column_name}::{col_type}") #FIXME not working
        except Exception as err:
            print("update_column err -> ",err)
        finally:
            conn.close()

    def create_table(self, table_name: str, *columns) -> bool:
        """ create prepared table with alembic """
        try:
            conn = self.engine.connect()

            # context config for alembic
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            op.create_table(
                table_name, *columns,
            )
            print(f"{table_name} is created")
            return True
        except Exception as error:
            print("_create_table err ->", error)
        finally:
            conn.close()

    def add_column(self, table_name: str, *column) -> bool:
        """ Add column to given table """
        try:
            conn = self.engine.connect()
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            for col in column:
                op.add_column(table_name, col)
            return True
        except Exception as err:
            print("add_column err -> ",err)
            return False
        finally:
            conn.close()

    def create_fk_constraint(self,fk_constraints:list,const_columns:dict) -> bool:
        """ Get list of foreign keys from static list `fk_constraints` and created it  """
        try:
            conn = self.engine.connect()
            ctx = MigrationContext.configure(conn)
            op = Operations(ctx)
            for constraint in fk_constraints:
                dest_table_name = constraint.pop("table_name")
                column_name = constraint.pop("column_name")
                source_table = constraint.pop("source_table")
                dest_column = constraint.pop("dest_column")
                temp = [i for i in const_columns[source_table]]
                if not dest_column in temp:
                    op.create_foreign_key(
                        None,
                        source_table,
                        dest_table_name,
                        [dest_column],
                        [column_name],
                        **constraint,
                    )
            return True
        except Exception as err:
            print("create_fk_constraint err -> ",err)
            return False
        finally:
            conn.close()

    def drop_fk(self, fk_constraints: str):
        try:
            conn = self.engine.connect()
            transactional = conn.begin()
            
            for fk in fk_constraints:
                conn.execute(DropConstraint(fk,cascade=True)) # maybe set cascade=True in future

            transactional.commit()
        except Exception as err:
            print("fk_drop -> ",err)
        finally:
            conn.close()

    def db_drop_everything(self,table_list):
        """ From http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DropEverything """
        try:
            conn = self.engine.connect()
            transactional = conn.begin()
            inspector = reflection.Inspector.from_engine(self.engine)

            tables = []
            all_foreign_keys = []

            for table_name in inspector.get_table_names():
                if table_name in table_list:
                    fks = []
                    for fk in inspector.get_foreign_keys(table_name):
                        if not fk["name"]:
                            continue
                        fks.append(ForeignKeyConstraint((), (), name=fk["name"]))
                    t = Table(table_name, self.metadata, *fks)
                    tables.append(t)
                    all_foreign_keys.extend(fks)

            for foreignkey in all_foreign_keys:
                conn.execute(DropConstraint(foreignkey))

            for table in tables:
                conn.execute(DropTable(table))

            transactional.commit()
        except Exception as err:
            print("err -> ",err)
            return False
        finally:
            conn.close()






__all__ = (
    "TableExists"
    "FileDoesNotExists"

)


class TableExists(Exception):
    def __init__(self, message,errors):
      
        self.message = message
        self.errors = errors


class FileDoesNotExists(Exception):
    def __init__(self, message):
      
        self.message = message



from madmigration.config.config_schema import MigrationTablesSchema
from madmigration.config.config_schema import ColumnParametersSchema
from madmigration.config.config_schema import TablesInfo
from sqlalchemy import Column, Table, MetaData, ForeignKey, ForeignKeyConstraint
from sqlalchemy.engine import reflection
from madmigration.errors import TableExists
from sqlalchemy.ext.mutable import MutableDict
from collections import defaultdict
from madmigration.postgresqldb.type_convert import get_type_object
from madmigration.config.conf import Config
from madmigration.db_operations.operations import DbOperations

from psycopg2 import errors
from sqlalchemy import exc

from queue import Queue


class BaseMigrate():
    q = Queue()
    def __init__(self, config: Config,destination_db):
        self.global_config = config
        self.migration_tables = config.migrationTables
        self.engine = destination_db.engine
        self.connection = destination_db
        self.metadata = MetaData()
        self.table_list = set()
        self.table_create = defaultdict(list)
        self.table_update = defaultdict(list)
        self.alter_col = defaultdict(list)
        self.fk_constraints = []
        self.dest_fk = []
        self.contraints_columns = defaultdict(set)
        self.db_operations = DbOperations(self.engine)

    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.engine.session.close()

    def collect_table_names(self):
        """ collects all tables that the program should create """
        try:
            for migrate_table in self.migration_tables:
                tabel_name = migrate_table.migrationTable.DestinationTable.name
                self.table_list.add(tabel_name)
        except Exception as err:
            print("err -> ",err)

    def collect_drop_fk(self):
        """ 
        Collect foreign key constraints for tables so 
        that if something goes wrong, delete everything
        """
        try:
            conn = self.engine.connect()
            transactional = conn.begin()
            inspector = reflection.Inspector.from_engine(self.engine)

            for table_name in inspector.get_table_names():
                if table_name in self.table_list:
                    for fk in inspector.get_foreign_keys(table_name):
                        if not fk["name"]:
                            continue
                        self.dest_fk.append(ForeignKeyConstraint((), (), name=fk["name"]))
                        self.contraints_columns[table_name].add(*fk["constrained_columns"])
            transactional.commit()
        except Exception as err:
            print("err -> ",err)
            return False
        finally:
            conn.close()
    

    def parse_migration_tables(self,tabels_schema:MigrationTablesSchema):
        """
        This function parses migrationTables from yaml file
        """
        try:
            self.source_table = tabels_schema.migrationTable.SourceTable.dict()
            self.destination_table = tabels_schema.migrationTable.DestinationTable.dict()
            self.columns = tabels_schema.migrationTable.MigrationColumns
        except Exception as err:
            print("err -> ",err)


    def parse_migration_columns(
        self, tablename: str, migration_columns: ColumnParametersSchema
    ):
        """
        This function parses migrationColumns schema and prepare column
        """
        try:
            # table_name, column_name,type_=col_type, **options
            update = self.check_table(tablename)
                
            for col in migration_columns:
                self.source_column = col.sourceColumn
                self.destination_column = col.destinationColumn
                self.dest_options = col.destinationColumn.options.dict()

                self._parse_fk(tablename, self.dest_options.pop("foreign_key"))
                column_type = self._parse_column_type()

                col = Column(self.destination_column.name, column_type, **self.dest_options)
                if update:
                    if not self.check_column(tablename, self.destination_column.name):
                    #     self.add_alter_column(tablename, {"column_name": self.destination_column.name,"type":column_type,"options":{**self.dest_options}})
                    # else:
                        self.add_updated_table(tablename,col)
                else:
                    self.add_created_table(tablename,col)
        except Exception as err:
            print("parse_migration_columns err -> ",err)


    def add_updated_table(self,table_name: str, col: Column):
        self.table_update[table_name].append(col)

    def add_created_table(self,table_name: str, col: Column):
        self.table_create[table_name].append(col)

    def add_alter_column(self,table_name: str, col: Column):
        self.alter_col[table_name].append(col)

    def prepare_tables(self):
        try:
            for migrate_table in self.migration_tables:
                if migrate_table.migrationTable.DestinationTable.create:
                    self.parse_migration_tables(migrate_table)
                    self.parse_migration_columns(self.destination_table.get("name"),self.columns)
        except Exception as err:
            print("prepare_tables -> ",err)

    def update_table(self):
        
        for tab,col in self.table_update.items():
            self.db_operations.add_column(tab,*col)
        return True
    
    def alter_columns(self):
        for tab, val in self.alter_col.items():
            for i in val:
                self.db_operations.update_column(tab,i.pop("column_name"),i.pop("type"), **i.pop("options"))
        return True
    
    def create_tables(self):
        for tab, col in self.table_create.items():
            self.db_operations.create_table(tab,*col)
        return True

    def process(self):
        """
        Create and check existing tables. 
        Collect foreign key constraints
        """
        try:
            # self.alter_columns()
            self.update_table()
            self.create_tables()
            self.db_operations.create_fk_constraint(self.fk_constraints,self.contraints_columns)
            return True
        except Exception as err:
            print("create_tables err -> ", err)

    def _parse_column_type(self) -> object:
        """ Parse column type and options (length,type and etc.) """
        
        try:
            column_type = self.get_column_type(self.dest_options.pop("type_cast"))
            type_length = self.dest_options.pop("length")
            if type_length:
                column_type = column_type(type_length)
            return column_type
        except Exception as err:
            print(err)
        
        # print(self.dest_options.get("length"))
        type_length = self.dest_options.pop("length")
        if type_length:
            column_type = column_type(type_length)
        return column_type

    def _parse_fk(self, tablename, fk_options):
        """ Parse foreignkey and options (use_alter,colum and etc.) """
        try:
            if fk_options:
                fk_options["source_table"] = tablename
                fk_options["dest_column"] = self.destination_column.name
                self.fk_constraints.append(fk_options)
        except Exception as err:
            print("_parse_fk err -> ", err)


    def check_table(self, table_name: str) -> bool:
        """ Check table exist or not, and wait user input """
        try:
            if self.engine.dialect.has_table(self.engine.connect(), table_name):
                return self.get_input(table_name)
            return False
        except Exception as err:
            print(err)
            return False

    def get_input(self,table_name):
        while True:
            answ = input(
                f"Table with name '{table_name}' already exist,\
'{table_name}' table will be dropped and recreated,your table data will be lost,process?(y/n) ")
            if answ.lower() == "y":
                self.db_operations.drop_fk(self.dest_fk)
                self.db_operations.drop_table(table_name)
                return False
            elif answ.lower() == "n":
                return True
            else:
                continue
        
    def get_table_attribute_from_base_class(self, source_table_name: str):
        """
        This function gets table name attribute from sourceDB.base.classes. Example sourceDB.base.class.(table name)
        Using this attribute we can query table using sourceDB.session
        :return table attribute
        """
        return getattr(self.connection.base.classes, source_table_name)

    def get_data_from_source_table(self, source_table_name: str, source_columns: list):

        table = self.get_table_attribute_from_base_class(source_table_name.name)
        rows = self.connection.session.query(table).yield_per(1)

        for row in rows:
            data = {}
            for column in source_columns:
                data[column] = getattr(row, column)
            yield data


    def check_column(self, table_name: str, column_name: str) -> bool:
        """
            Check column exist in destination table or not
            param:: column_name -> is destination column name
        """
        try:
            insp = reflection.Inspector.from_engine(self.engine)
            has_column = False
            for col in insp.get_columns(table_name):
                if column_name not in col["name"]:
                    continue
                return True
            return has_column
        except Exception as err:
            print("check_column err -> ",err)
            return False


    @staticmethod
    def insert_data(engine, table_name, data: dict):
        # stmt = None
        try:
            stmt = engine.base.metadata.tables[table_name].insert().values(**data)
           
        except Exception as err:
            print("stmt err: ", err)
            return
        # print("STMT",stmt)
  
        try:
            engine.session.execute(stmt)
        except Exception as err:
            print("Passing stmt into queue")
            BaseMigrate.put_queue(stmt)

        try:
            engine.session.commit()
        except Exception as err:
            print("Commit error: ", err)
            engine.session.rollback()
        finally:
            engine.session.close()

    @staticmethod
    def insert_queue(engine):
        for stmt in BaseMigrate.q.queue:
            try:
                print("Inserting from queue: ")
                engine.session.execute(stmt)
            except Exception as err:
                print("insert_queue err: ", err)
            
            try:
                engine.session.commit()
            except Exception as err:
                print("insert_queue Commit error: ", err)
                engine.session.rollback()
            finally:
                engine.session.close()


    @staticmethod
    def put_queue(data):
        BaseMigrate.q.put(data)

    @staticmethod
    def get_queue():
        return BaseMigrate.q.get()

    @staticmethod
    def type_cast(data_from_source, mt, convert_info: dict):
        for columns in mt.migrationTable.MigrationColumns:
            source_column = columns.sourceColumn.name
            destination_column = columns.destinationColumn.name
            
            if columns.destinationColumn.options.type_cast:
                destination_type_cast = columns.destinationColumn.options.type_cast
            else:
                destination_type_cast = None

            if convert_info.get(destination_column):
                # ClassType is Class of data type (int, str, float, etc...)
                # Using this ClassType we are converting data into format specified in type_cast
                datatype = get_type_object(destination_type_cast)               

                try:
                    if datatype == type(data_from_source.get(source_column)):
                        data_from_source[destination_column] = data_from_source.pop(source_column)

                    elif datatype.__name__ == "UUID":
                        data_from_source[destination_column] = str(data_from_source.pop(source_column))
                    else:
                        try:
                            data_from_source[destination_column] = datatype(data_from_source.pop(source_column))
                        except Exception as err:
                            data_from_source[destination_column] = None
                       
                except Exception as err:
                    data_from_source[destination_column] = None
            else:
                data_from_source[destination_column] = data_from_source.pop(source_column)
            
        return data_from_source


    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type
        :param type_name: str
        :return: object class
        """
        raise NotImplementedError





from madmigration.db_init.connection_engine import SourceDB
from madmigration.db_init.connection_engine import DestinationDB
from madmigration.utils.helpers import detect_driver, get_cast_type
from sqlalchemy import Column, MetaData, Table
from madmigration.config.conf import Config
from madmigration.mysqldb.type_convert import get_type_object
from sqlalchemy import insert
from sqlalchemy import (
    Integer,
    String,
    DateTime,
    Date,
    BigInteger,
    VARCHAR,
    Float,
    TIMESTAMP
)
from time import sleep


class Controller:
    def __init__(self, migration_config: Config):
        self.config = migration_config

        # Source and Destination DB Initialization with session
        self.source_db = SourceDB(self.config.source_uri)
        self.destination_db = DestinationDB(self.config.destination_uri)
        self.metadata = MetaData()
        # Source and Destination Database - all tables (NOT MIGRATION TABLES!!!)
        self.sourcedb_all_tables_names = self.source_db.engine.table_names()
        self.destinationdb_all_tables_names = self.destination_db.engine.table_names()

        # All migration tables (Yaml file migrationTables)
        self.migration_tables = self.config.migrationTables

        # Destination DB Driver and name
        self.destinationdb_driver = self.destination_db.engine.driver
        self.destinationdb_name = self.destination_db.engine.name

        # Source DB Driver
        self.sourcedb_driver = self.source_db.engine.driver

    

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.source_db.session.close()
        self.destination_db.session.close()
        return True # we take care of error handling, wrap it up.

    

    def run_table_migrations(self):
        # detect migration class
        try:
            
            migrate = detect_driver(self.destinationdb_driver)
            
            mig = migrate(self.config,self.destination_db)
            mig.prepare_tables()
            mig.process()
        except Exception as err:
            print(err)
            
        
        # migrate.create_fk_constraint(self.destination_db.engine)

    def get_column_type_from_source_table(self, table_name, column_name):
        """
        Get type of column in source table from Source Database
        :param table_name: Table name
        :param column_name: Column name
        :return: Column type (Integer, Varchar, Float, Double, etc...)
        """
        tables = self.source_db.base.metadata.tables
        table = tables.get(table_name)
        for column in table.columns:
            if column.name == column_name:
                return column.type

    def run(self):
        
        # Looping in migrationTables
        self.destination_db.base.prepare(self.destination_db.engine, reflect=True)

        for mt in self.migration_tables:
            self.destination_table = mt.migrationTable.DestinationTable

            # Migrate class is use based on driver
            Migrate = detect_driver(self.sourcedb_driver)
            
            # Instance of Migrate class
            migrate = Migrate(self.config, self.source_db)
            migrate.parse_migration_tables(mt)    

            # Dictionary is used to keep data about destination column and type_cast format
            self.convert_info = {}

            # Columns List is used to keep source Columns names
            # We will send table name and source columns list to function "get_data_from_source_table"
            # get_data_from_source_table function will yield data with specified columns from row
            columns = []
            
            for column in migrate.columns:
                columns.append(column.sourceColumn.name)

                if column.destinationColumn.options.type_cast:
                    self.convert_info[
                        column.destinationColumn.name] = column.destinationColumn.options.type_cast
            
            # self.source_data is data received (yield) from get_data_from_source_table function
            self.source_data = migrate.get_data_from_source_table(mt.migrationTable.SourceTable, columns)
            
            for source_data in self.source_data:
                try:
                    new_data = Migrate.type_cast(data_from_source=source_data, mt=mt, convert_info=self.convert_info)
                except Exception as err:
                    print("Migrate.type_cast -> ", err)
                try:
                    Migrate.insert_data(engine=self.destination_db, table_name=self.destination_table.name, data=new_data)
                except Exception as err:
                    print("Migrate.insert_data -> ",err)
        
        # Queue is used to hold data that could not be inserted with foreignkey error. Loop from queue again to insert data with fk
        Migrate.insert_queue(self.destination_db)


                
















import yaml
from madmigration.config.config_schema import ConfigSchema
from pprint import pprint


# Config class generates configuration based on yaml file
class Config:
    def __init__(self, config_yaml):
        self.config_file = config_yaml

        with open(self.config_file) as f:
            self.config_data = ConfigSchema(**yaml.load(f, Loader=yaml.FullLoader)) # noqa  E501
        # pprint(self.config_data.dict())
        self.version = self.config_data.version
        self.source_uri = self.config_data.Configs[0].SourceConfig.get("dbURI")
        self.destination_uri = self.config_data.Configs[1].DestinationConfig.get("dbURI") # noqa  E501
        self.migrationTables = self.config_data.migrationTables





from datetime import datetime
from sqlalchemy.schema import DropTable
from sqlalchemy.schema import ForeignKeyConstraint
from sqlalchemy.ext.compiler import compiles
from typing import Union
import os
from pathlib import Path
from sqlalchemy import (
    Integer,
    String,
    DateTime,
    Date,
    BigInteger,
    VARCHAR,
    Float,
    TIMESTAMP
)
from madmigration.errors import FileDoesNotExists

from madmigration.mysqldb.migration import MysqlMigrate
from madmigration.postgresqldb.migration import PgMigrate
from madmigration.mssql.migration import MssqlMigrate

###########################
# Get class of cast #
###########################
def get_cast_type(type_name: str) -> object:
    """
    :param type_name: str
    :return: object class
    """
    return {
        'str': str,
        'string': str,
        'int': int,
        'integer': int,
        'float': float,
        'datetime': datetime,
        'varchar': VARCHAR
    }.get(type_name.lower())


###########################
# Detect db driver fro migration #
###########################
def detect_driver(driver: str) -> Union[MysqlMigrate, PgMigrate]:
    """
    :param driver: str
    :return: object class
    """
    return {
        "mysqldb" : MysqlMigrate,
        "mysql+mysqldb": MysqlMigrate,
        "pymysql": MysqlMigrate,
        "mysql+pymysql" : MysqlMigrate,
        "mariadb+pymsql" : MysqlMigrate,
        "psycopg2": PgMigrate,  
        "pg8000": PgMigrate,
        "pyodbc": MssqlMigrate
       # "postgresql+asyncpg": postgres_migrate,
       # "asyncpg": postgres_migrate
    }.get(driver)


@compiles(DropTable, "postgresql")
def _compile_drop_table(element, compiler, **kwargs):
    """
    Drop table cascade with foreignkeys
    https://stackoverflow.com/questions/38678336/sqlalchemy-how-to-implement-drop-table-cascade# 
    """
    return compiler.visit_drop_table(element) + " CASCADE"



@compiles(ForeignKeyConstraint, "mysql", "mariadb")
def process(element, compiler, **kw):
    element.deferrable = element.initially = None
    return compiler.visit_foreign_key_constraint(element, **kw)


def check_file(file):
    return Path(file).is_file() and os.access(file, os.R_OK)

def file_not_found(file):
    raise FileDoesNotExists(f"Given file does not exists file: {file}",)


def issue_url():
    return "https://github.com/MadeByMads/mad-migration/issues"

def app_name():
    return "madmigrate"


def parse_uri(uri):
    if "///" in uri:
        database_name  = uri.split("///")[-1]
    else:
        database_name = uri.split("/")[-1]

    return database_name


from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.dialects.mssql import (
    BIGINT,
    BINARY,
    BIT,
    CHAR,
    DATE,
    DATETIME,
    DATETIME2,
    DATETIMEOFFSET,
    DECIMAL,
    FLOAT,
    IMAGE,
    INTEGER,
    MONEY,
    NCHAR,
    NTEXT,
    NUMERIC,
    NVARCHAR,
    REAL,
    SMALLDATETIME,
    SMALLINT,
    SMALLMONEY,
    TEXT,
    TIME,
    TIMESTAMP,
    TINYINT,
    UNIQUEIDENTIFIER,
    VARBINARY,
    VARCHAR,
)
from sqlalchemy import DateTime
from sqlalchemy_utils import UUIDType
from madmigration.basemigration.base import BaseMigrate
from madmigration.config.conf import Config
from pprint import pprint

#TODO for fk keys RESTRICT|CASCADE|SET NULL|NO ACTION|SET DEFAULT

class MssqlMigrate(BaseMigrate):
    def __init__(self, config: Config, destination_db):
        super().__init__(config, destination_db)
        self.collect_table_names()
        self.collect_drop_fk()

    @staticmethod
    def get_column_type(type_name: str) -> object:
        """Get class of db type
        :param type_name: str
        :return: object class
        """
        return {
            "varchar": VARCHAR,
            "char": CHAR,
            "string": VARCHAR,
            "text": TEXT,
            "integer": INTEGER,
            "smallint": SMALLINT,
            "bigint": BIGINT,
            "binary": BINARY,
            "boolean": BIT, #FIXME  raise error when insert bool value
            "bool": BIT,   #FIXME  raise error when insert bool value
            "bit": BIT,
            "date": DATE,
            "datetime": DATETIME,
            "timestamp": TIMESTAMP,
            "time": TIME,
            "float": FLOAT,
            "real": REAL,
            # "array": ARRAY, #FIXME column with array include array elemet type argument
            "numeric": NUMERIC,
            "money": MONEY,
            "uuid": UUIDType(binary=False),
        }.get(type_name.lower())




import os
import pytest
from madmigration.config.config_schema import ConfigSchema
from madmigration.config.conf import Config
import sys
import yaml


def config_schema_test():

    file = os.path.realpath('mysql.yaml')
    obj = Config(file)

    assert obj.version == obj.config_data.version
    assert obj.source_uri == obj.config_data.Configs[0].SourceConfig.get("dbURI")
    assert obj.destination_uri == obj.config_data.Configs[1].DestinationConfig.get("dbURI")
    assert obj.migrationTables == obj.config_data.migrationTables
    
    assert obj.config_file == file





from datetime import datetime, date
from time import time
from enum import Enum
import json
from uuid import UUID

def get_type_object(data_type):
    """
    :param data_type: data type
    :return: object class
    """
    return {
        "varchar": str,
        "char": str,
        "string": str,
        "text": str,
        "integer": int,
        "smallint": int,
        "bigint": int,
        "binary": bytes,
        "boolean": bool, 
        "bool": bool,  
        "bit": int,
        "date": date,
        "datetime": datetime,
        "timestamp": datetime,
        "time": datetime,
        "float": float,
        "real": float,
        "numeric": float,
        "money": float,
        "uuid": UUID,
        None: None

    }.get(data_type.lower())



from setuptools import setup, find_packages
import madmigration

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="madmigration",
    packages=find_packages(),
    version="0.1.6",
    entry_points="""
        [console_scripts]
        madmigrate=madmigration.scripts.commands:cli
    """,
    author="Hasan Aliyev, Tural Muradov, Sabuhi Shukurov",
    author_email="hasan.aliyev.555@gmail.com, tural_m@hotmail.com, sabuhi.shukurov@gmail.com",
    license='MIT',
    description="Mad migration",
    long_description_content_type="text/markdown",
    url="https://github.com/MadeByMads/mad-migration",
    long_description=long_description,
    install_requires=["click>=7.1.2", "SQLAlchemy-Utils>=0.36.8","SQLAlchemy>=1.3.18",'pydantic>=1.7.2',"alembic>=1.4.2", "PyYAML>=5.3.1"],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_packge_data=True,
    tests_require=[
        'pytest',
        'mock',
    ],
    platforms=['any'],
    python_requires='>=3.6',
    project_urls={  
        'Bug Reports': 'https://github.com/MadeByMads/mad-migration/issues',
        'Say Thanks!': 'https://github.com/MadeByMads/mad-migration/network/dependencies',
        'Source': 'https://github.com/MadeByMads/mad-migration',
    },
    extras_require={
        'mysql': ["mysqlclient>=2.0.1"],
        'postgresql': ["psycopg2>=2.8.5"],
    },
)






from datetime import datetime, date
from enum import Enum
import json
from uuid import uuid4, UUID

def get_type_object(data_type):
    """
    :param data_type: data type
    :return: object class
    """
    return {
        "uuid": UUID,
        "string": str,
        "str": str,
        "varchar": str,
        "text": str,
        "nvarchar": str,
        "smallint": str,
        "char": str,

        "int": int,
        "integer": int,
        "bigint": int,

        "float": float,
        "numeric": float,
        "decimal": float,

        "date": date,
        "datetime": datetime,

        "binary": bytes,
        "enum": Enum,
        "set": set,

        "json": dict,

        "boolean": bool,
        "bool": bool,
        None: None

    }.get(data_type.lower())



import click
import math
import random
import time
import sys
from madmigration.config.conf import  Config
from madmigration.main import Controller
from madmigration.utils.helpers import check_file, file_not_found

@click.group()
def cli():
    """Controller group Cli"""
    

@cli.command(help='simple Migrate ready on hand with CLI')
@click.option('--file',"-f",metavar='YAML file',prompt='YAML file',show_default=True,required=True, help='YAML file')

#TODO valdait yaml file if given file is correct, drop fk.
def cli(file):  

    if check_file(file):
        config = Config(file)
        with Controller(config) as app:

            app.run_table_migrations()
            # app.run()
    else:
        file_not_found(file)

    